<?php
/* CurrencyExchange class
 * Version 1.0.1
 * Copyright 2004-2006, Steve Blinch
 * http://code.blitzaffe.com
 *
 * Returns a list of exchange rates for various currencies.
 *
 *
 * WARNING
 * 
 * The values and dates provided by this script are believed to be reliable,
 * but the author makes no warranties regarding this code, the values it 
 * provides, or their fitness for a particular purpose, accuracy or 
 * availability.
 * 
 * Note that this script relies entirely on several third-party websites 
 * (such as the Federal Reserve Bank of New York, the Bank of Canada, and
 * so-on) to provide exchange rate information; if those sites become 
 * unreachable for any reason, this script will be unable to obtain exchange
 * rate information.
 * 
 * It is *strongly* recommended that you use the values returned by this 
 * script for informational purposes only (to provide an estimate to your
 * customers).  You *SHOULD NOT* rely on this script for unattended financial
 * transactions, as there are several situations (all of which are beyond the
 * control of this script) which could result in incorrect exchange rates.
 * 
 * 
 * CHANGELOG
 *
 * 1.0.1 (10-Dec-2006)
 *			- Rewrote Federal Reserve Bank of New York parser to handle
 *			  FRBNY's new XML format.
 * 
 * 
 * EXAMPLE
 * 
 * // CurrencyExchange class usage example
 * require_once('class_CurrencyExchange.php');
 * $exchange = new CurrencyExchange();
 * if ($exchange->retrieve()) {
 * 	
 * 	// Display the current Canadian exchange rate information
 * 	var_dump($exchange->rates["CAD"]);
 * 	
 * 	// Display all exchange rates returned by the CurrencyExchange class
 * 	foreach ($exchange->rates as $currency=>$rate) {
 * 		echo "Currency: $currency<br>".
 * 			"Exchange rate: ".	$rate["rate"]."<br>".
 * 			"Country: ".		$rate["unit"]."<br>".
 * 			"Units: ".			$rate["unit"]."<br>".
 * 			"Date retrieved: ".	$rate["date"]."<br><br>";
 * 	}
 * 	
 * 	// Display information about the source of the exchange rate information
 * 	// (eg: if it was retrieved from the Federal Reserve Bank of NY, the
 * 	// Bank of Canada, etc.)
 * 	var_dump($exchange->source);
 * }
 * // ------------------------------------------------------------------------
 * 
 * 
 * 
 * CACHING
 * 
 * It is strongly recommended that you cache the exchange rates returned by this
 * script to avoid overloading the source servers (and to reduce the probability
 * of a failure due to a temporary source server outage).
 * 
 * The CurrencyExchange class has a built-in serializing system that determines
 * whether or not the cached data is expired.  All you have to do is store the
 * data in between requests, and provide it to the CurrencyExchange class on the
 * next load.  For example:
 * 
 * // ------------------------------------------------------------------------
 * 
 * // You need to implement the load_cached_rates() function, which should retrieve
 * // the cached rates (from a file, or database, etc.)
 * $cached_rates = load_cached_rates();
 * 
 * $exchange = new CurrencyExchange();
 * 
 * // Use the set_cached() method to pass the cached data back to the class.
 * $exchange->set_cached($cached_rates);
 * 
 * // The retrieve() method will automatically detect whether or not the
 * // rates in $cached_rates are expired; if so, it will retrieve new rates,
 * // otherwise it will reuse the cached rates (and not contact the server).
 * if ($exchange->retrieve()) {
 * 	
 * 	// Display the current Canadian exchange rate information
 * 	var_dump($exchange->rates["CAD"]);
 * 	
 * 	// Display all exchange rates returned by the CurrencyExchange class
 * 	foreach ($exchange->rates as $currency=>$rate) {
 * 		echo "Currency: $currency<br>".
 * 			"Exchange rate: ".	$rate["rate"]."<br>".
 * 			"Country: ".		$rate["country"]."<br>".
 * 			"Units: ".			$rate["unit"]."<br>".
 * 			"Date retrieved: ".	$rate["date"]."<br><br>";
 * 	}
 * 	
 * 	// Display information about the source of the exchange rate information
 * 	// (eg: if it was retrieved from the Federal Reserve Bank of NY, the
 * 	// Bank of Canada, etc.)
 * 	var_dump($exchange->source);
 * 	
 * 	// Use the get_cached() method to retrieve a string containing the
 * 	// exchange rate information encoded for caching.
 * 	$new_cached_rates = $exchange->get_cached();
 * 	
 * 	// You need to implement the save_cached_rates() function, which should
 * 	// store the cached rates (to a file, or database, etc.)
 * 	save_cached_rates($new_cached_rates);
 * }
 * // ------------------------------------------------------------------------
 * 
 */


class CurrencyExchange {

	function CurrencyExchange() {
	
		// List of exchange rate sources.
		//
		// This class tries each rate source in sequence, so you should order your sources from most
		// reliable to least reliable.
		// 
		$this->sources = array(
			// Probably the most reliable place to get your rates from
			"fedreserveny"=>array(
				"name"=>"Federal Reserve Bank of New York",
				"url"=>"http://www.ny.frb.org/markets/fxrates/FXtoXML.cfm?FEXdate=%Y%2D%m%2D%d%2000%3A00%3A00%2E0&FEXtime=1200",				
				"sourcepage"=>"http://www.ny.frb.org/markets/fxrates/noon.cfm",
				"notes"=>"These Daily 12 noon buying rates in New York are certified by the New York Federal Reserve Bank for customs purposes.",
				"parser"=>"fedreserveny"
			),
/*
			// Fallback in case the NY Federal Reserve Bank's site is ever unreachable
			"bankofcanada"=>array(
				"name"=>"Bank of Canada CSV",
				"url"=>"http://www.bankofcanada.ca/en/financial_markets/csv/exchange_eng.csv",
				"sourcepage"=>"http://www.bankofcanada.ca/en/exchange.htm",
				"notes"=>"These rates are nominal quotations  neither buying nor selling rates  and are intended for statistical or analytical purposes. Rates available from financial institutions will differ.",
				"parser"=>"bankofcanadacsv"
			),
			
			// The CSV link above was marked "Subject to Change", so we're include the HTML link
			// as well as an additional fallback
			"bankofcanadahtml"=>array( 
				"name"=>"Bank of Canada HTML",
				"url"=>"http://www.bankofcanada.ca/fmd/exchange.htm",
				"sourcepage"=>"http://www.bankofcanada.ca/en/exchange.htm",
				"notes"=>"These rates are nominal quotations  neither buying nor selling rates  and are intended for statistical or analytical purposes. Rates available from financial institutions will differ.",
				"parser"=>"bankofcanadahtml"
			)
*/			
		);
		
		
		// Different sources refer to countries in different ways.  For example, the
		// Federal Reserve Bank of New York refers to "Australia", whereas the Bank of
		// Canada refers to it as "Australian" (note the trailing "n").
		//
		// This array serves as a translation table to determine which currency belongs
		// to which country, taking into account the variations in the country names.
		//
		// The name of each country (as retrieved from the remote server) is also replaced
		// with the FIRST country name in the list (eg: if the server returned "Canadian",
		// and "Canada" was the first spelling entered below for CAD, then "Canada" would
		// be returned as the country name).
		//
		// Entries are not case-sensitive.
		$this->countrytranslation = array(
			"AUD"=>array("Australia","Australian"),
			"EUR"=>array("Europe","European","European Monetary Union","European Union","Euro de l'Union montaire"),
			"USD"=>array("United States","USA","U.S."),
			"CAD"=>array("Canada","Canadian"),
			"GBP"=>array("United Kingdom","Pound"),
			"JPY"=>array("Japan","Japanese"),
			"CHF"=>array("Switzerland","Swiss"),
			"DKK"=>array("Denmark","Danish"),
			"HKD"=>array("Hong Kong"),
			"INR"=>array("India","Indian"),
			"KRW"=>array("South Korea","South Korean"),
			"MXN"=>array("Mexico","Mexican"),
			"MYR"=>array("Malaysia","Malaysian"),
			"NOK"=>array("Norway","Norwegian"),
			"NZD"=>array("New Zealand"),
			"SEK"=>array("Sweden","Swedish"),
			"SGD"=>array("Singapore"),
			"THB"=>array("Thailand","Thai"),
			"TWD"=>array("Taiwan","Taiwanese New"),
			"VEB"=>array("Venezuela","Venezuelan"),
			"ZAR"=>array("South Africann","South African")
		);
	}

	// Retrieves up-to-date exchange rate information.
	//
	// Returns TRUE if any one of the sources was successfully retrieved, 
	// FALSE if all sources failed.
	function retrieve() {
		// if we're using cached rates, return immediately
		if ($this->use_cached) return true;
		
		reset($this->sources);
		foreach ($this->sources as $name=>$source) {
			if ($this->process_source($source)) {
				return true;
			}
		}
		return false;
	}
	
	function retrieve_file($url) {
		// do some date arithmetic to find the last weekday if this is not
		// presently a weekday
		$d = date('d');
		switch(date('w')) {
			case 0: $d -= 2; break;
			case 6: $d--; break;
		}
		$urldate = mktime(0,0,0,date('m'),$d,date('Y'));
		
		$url = str_replace(
			array('%Y','%m','%d'),
			array(date('Y',$urldate),date('m',$urldate),date('d',$urldate)),
			$url
		);
		$fp = fopen($url,"r");
		if (!$fp) {
			return false;
		}
		$content = "";
		while (!feof($fp)) {
			$content .= fread($fp,1024);
		}
		fclose($fp);
		
		return $content;
	}
	
	function process_source($source) {
		$method_name = "parse_".$source["parser"];
		if (!method_exists($this,$method_name)) return false;

		$content = $this->retrieve_file($source["url"]);
		if (!$content) return false;
		
		$this->rates = $this->$method_name($content);
		
		if ($this->rates) {
			ksort($this->rates);
			$this->source = $source;
			return true;
		} else {
			return false;
		}
		
	}
	
	function get_country_currency($find_country,$find_unit) {
		if (is_null($find_country)) {
			$countries = $this->countrytranslation[$find_unit];
			return array($countries[0],$find_unit);
		}
		$find_country = strtolower($find_country);

		reset($this->countrytranslation);
		foreach ($this->countrytranslation as $currency=>$countries) {
			foreach ($countries as $k=>$country) {
				if (strtolower($country)==$find_country) {
					return array($countries[0],$currency);
				}
			}
		}
		return array(false,false);
	}
	
	function parse_fedreserveny($content) {
		$rates = false;
		
		if (!preg_match_all("/<frbny:Series (.*?)>([\s\S]*?)<\/frbny:Series>/i",$content,$matches,PREG_SET_ORDER)) return false;
		foreach ($matches as $k=>$match) {
			list(,$rateinfo,$elements) = $match;	
			
			if (preg_match("/UNIT=\"(.*?)\"/i",$rateinfo,$match)) {
				$unit = trim($match[1]);
				if (!$unit) continue;
			} else {
				continue;
			}
			
			if (preg_match("/<frbny:TIME_PERIOD>(.*?)<\/frbny:TIME_PERIOD>/i",$elements,$match)) $date = $match[1];
			if (preg_match("/<frbny:CURR>(.*?)<\/frbny:CURR>/i",$elements,$match)) $inusd = ($match[1]=='USD');
			if (preg_match("/<country unitname=\"(.*?)\">(.*?)<\/country>/i",$elements,$match)) list(,$unitname,$country) = $match;
			if (preg_match("/<frbny:OBS_VALUE>([0-9\.]+)<\/frbny:OBS_VALUE>/i",$elements,$match)) $exchangerate = $match[1];

			$exchangerate = (float) $exchangerate;
						
			if ($inusd) {
				$exchangerate = round(1 / $exchangerate,4);
			}
			
			list($country,$currency) = $this->get_country_currency(NULL,$unit);
			
			// don't add it to the list if we don't have an exchange rate and currency
			if (!$exchangerate || !$currency) continue;
			
			$rates[$currency] = array(
				"date"=>$date,
				"unit"=>$unit,
				"country"=>$country,
				"rate"=>$exchangerate
			);
		}

		return $rates;
	}
	
	
	function fix_banqueducanada_date($date) {
		if (!preg_match("/([0-9]{2})\/([0-9]{2})\/([0-9]{4})/",$date,$match)) return $date;
		
		list(,$month,$day,$year) = $match;

		return "{$year}-{$month}-{$day}";		
	}
	
	function get_banqueducanada_nationinfo($nation) {
		if (preg_match("/^(.*)\s(.*?)\s\((.*?)\)$/",$nation,$match)) {
			$country = $match[1];
			$unitname = $match[2] .= " (".$match[3].")";
		} elseif (preg_match("/^(.*)\s(.*?)$/",$nation,$match)) {
			$country = $match[1];
			$unitname = $match[2];
		} else {
			$country = $unitname = false;
		}
		return array($country,$unitname);
	}
	
	function parse_bankofcanadacsv($content) {
		$rates = false;
		$dates = array();
		
		$matches = explode("\n",$content);
		foreach ($matches as $k=>$match) {
			$match = trim($match);
			if ((!$match) || (substr($match,0,1)=="#")) continue;

			$values = explode(",",$match);
			$nation = array_shift($values);
			if (strtolower(substr($nation,0,5))=="date ") {
				$date = $this->fix_banqueducanada_date($values[count($values)-1]);
				continue;
			}
			
			$exchangerate = (float) $values[count($values)-1];

			list($country,$unitname) = $this->get_banqueducanada_nationinfo($nation);			
			list($country,$currency) = $this->get_country_currency($country);

			// don't add it to the list if we don't have an exchange rate and currency
			if (!$exchangerate || !$currency) continue;
			
			$rates[$currency] = array(
				"date"=>$date,
				"unit"=>$unitname,
				"country"=>$country,
				"rate"=>1/$exchangerate
			);
		}

		// Bank of Canada returns its rates in Canadian dollars; convert to U.S. dollars
		// for standardization, and add a CAD record.

		$usrate = $rates["USD"]["rate"];
		if (!$usrate) return false;
		
		$usexchange = 1 / $usrate;
		unset($rates["USD"]);

		foreach ($rates as $currency=>$rate) {
			$rates[$currency]["rate"] = round($rates[$currency]["rate"] * $usexchange,4);
		}
		
		$rates["CAD"] = array(
			"date"=>$date,
			"unit"=>"Dollar",
			"country"=>"Canada",
			"rate"=>$usexchange
		);
		
		
		return $rates;
	}	
	
	function parse_bankofcanadahtml($content) {
		$rates = false;
		$dates = array();
		
		$lines = explode("\n",$content);
		foreach ($lines as $k=>$line) {
			$line = trim(strip_tags($line));
			if ((!$line) || (substr($line,0,1)=="#")) continue;
			
			if (preg_match("/^([0-9]{2})\/([0-9]{2})\/([0-9]{2})(\s+)([0-9]{2})\/([0-9]{2})\/([0-9]{2})(\s+)([0-9]{2})\/([0-9]{2})\/([0-9]{2})(\s+)([0-9]{2})\/([0-9]{2})\/([0-9]{2})(\s+)([0-9]{2})\/([0-9]{2})\/([0-9]{2})$/",$line,$match)) {
				$date = "20".$match[count($match)-3]."-".$match[count($match)-2]."-".$match[count($match)-1];
				continue;
			}
			
			
			if (preg_match("/^(.*?)(\s+)([0-9\.]+)(\s+)([0-9\.]+)(\s+)([0-9\.]+)(\s+)([0-9\.]+)(\s+)([0-9\.]+)(\s*)$/",$line,$match)) {
				$nationinfo = $match[1];
				$exchangerate = (float) $match[count($match)-2];
				list($nation,) = explode("/",$nationinfo);
				$nation = trim($nation);
			} else {
				continue;
			}

			list($country,$unitname) = $this->get_banqueducanada_nationinfo($nation);
			list($country,$currency) = $this->get_country_currency($country);

			// don't add it to the list if we don't have an exchange rate and currency
			if (!$exchangerate || !$currency) continue;
			
			$rates[$currency] = array(
				"date"=>$date,
				"unit"=>$unitname,
				"country"=>$country,
				"rate"=>1/$exchangerate
			);
		}

		// Bank of Canada returns its rates in Canadian dollars; convert to U.S. dollars
		// for standardization, and add a CAD record.

		$usrate = $rates["USD"]["rate"];
		if (!$usrate) return false;
		
		$usexchange = 1 / $usrate;
		unset($rates["USD"]);

		foreach ($rates as $currency=>$rate) {
			$rates[$currency]["rate"] = round($rates[$currency]["rate"] * $usexchange,4);
		}
		
		$rates["CAD"] = array(
			"date"=>$date,
			"unit"=>"Dollar",
			"country"=>"Canada",
			"rate"=>$usexchange
		);
		
		return $rates;
	}
	
	function set_cached($cached_rates) {
		$x = @base64_decode($cached_rates);
		if (!$x) return false;
		
		$rates = @unserialize($x);
		if (!is_array($rates)) return false;
		
		$cachedate = getdate($rates["date"]);
		$nowdate = getdate();
		
		$expired = 
			($nowdate["year"]!=$cachedate["year"]) ||
			($nowdate["mon"]!=$cachedate["mon"]);
		
		if (!$expired) {
			// last poll was this month of this year, so....
			
			// more than 1 day old? expired!
			if ($nowdate["mday"]>$cachedate["mday"]+1) $expired = true; 
			
			// retrieved before 1pm today?  expired!
			if (($nowdate["mday"]==$cachedate["mday"]) && ($cachedate["hours"]<=13) && ($nowdate["hours"]>13)) $expired = true;
	
			// retrieved before 1pm yesterday?  expired!
			if (($nowdate["mday"]==$cachedate["mday"]+1) && ($cachedate["hours"]<=13)) $expired = true;
			
			// retrieved after 1pm yesterday, and it's after 1pm today?  expired!
			if (($nowdate["mday"]==$cachedate["mday"]+1) && ($cachedate["hours"]>13) && ($nowdate["hours"]>13)) $expired = true;
		}
		
		if (!$expired) {
			$this->rates = $rates["rates"];
			$this->use_cached = true;
		} else {
			$this->use_cached = false;
		}
		return true;
	}

	function get_cached() {
		return
			base64_encode(
				serialize(
					array(
						"date"=>time(),
						"rates"=>$this->rates
					)
				)
			);
	}
	
}


?>